package com.miaozhen.device.support;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;

import com.miaozhen.device.IDCServer;
import com.miaozhen.device.ServerFactory;
import com.miaozhen.device.common.Constants;
import com.miaozhen.device.expection.DeviceException;
import com.miaozhen.device.response.PlaybackStatusResponse;
import com.miaozhen.device.response.Response;
import com.miaozhen.device.smi.PlaybackStatus;
import com.miaozhen.device.support.dolby.*;
import com.miaozhen.util.FileUtil;
import org.apache.commons.httpclient.params.HttpClientParams;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.io.SAXReader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.barco.dual.projector.EPlatform;
import com.barco.dual.projector.OSinfo;

public class DolbyServer implements IDCServer {

    HttpClientParams params = new HttpClientParams();

    protected SystemManagementDolby systemManagement;
    protected PlaybackControlDolby playbackControl;

    protected String ip;

    private final static String dolbyLogPath = "dolbyLog";

    private final Logger logger = LoggerFactory.getLogger(DolbyServer.class);

    private boolean isConnected = false;

    public DolbyServer(String ip, int port, String userName, String password) {
        this.ip = ip;
        isConnected = checkConnect();
        params.setParameter(HttpClientParams.USE_EXPECT_CONTINUE, Boolean.FALSE);
        // 设置ws连接超时时间
        params.setParameter(HttpClientParams.CONNECTION_MANAGER_TIMEOUT, 5000l);
        params.setParameter(HttpClientParams.SO_TIMEOUT, 5000);
        params.setParameter(HttpClientParams.HEAD_BODY_CHECK_TIMEOUT, 5000);
        playbackControl = new PlaybackControlDolby(params, ip);
        systemManagement = new SystemManagementDolby(params, ip);

    }

    @Override
    public void connect() throws DeviceException {
        isConnected = checkConnect();
    }

    @Override
    public boolean ejectSPL(String uuid) throws Exception {
        return false;
    }

    @Override
    public String getServerType() {
        return ServerFactory.ServerType.DOLBY.toString();
    }

    @Override
    public boolean getConnected(){
        return isConnected;
    }

    @Override
    public void disconnect() throws DeviceException {

    }

    public boolean checkConnect() {
        EPlatform OSInfo = OSinfo.getOSname();
        Runtime runtime = Runtime.getRuntime();
        Process process = null;
        try {
            if (OSInfo.equals(EPlatform.Windows)) {
                process = runtime.exec("ping " + this.ip + " -n 1 -w 3");
            } else {
                process = runtime.exec("ping " + this.ip + " -c 1 -w 3");
            }
            InputStream is = process.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);
            String line = null;
            StringBuffer sb = new StringBuffer();
            while ((line = br.readLine()) != null) {
                sb.append(line);
            }
            is.close();
            isr.close();
            br.close();
            if (null != sb && !sb.toString().equals("")) {
                if (sb.toString().toLowerCase().indexOf("ttl") > 0) {
                    return true;
                } else {
                    return false;
                }
            }
            process.destroy();
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        } finally {
            if (process != null) {
                process.destroy();
            }
        }
        return false;
    }

    @Override
    public void setConnected(boolean connected) {
        isConnected = connected;
    }

    @Override
    public String getIP() {
        return ip;
    }

    @Override
    public void setIP(String ip) {
        this.ip = ip;
    }

    @Override
    public String getAuditLog(String ftpIp, String user, String pass, String tempLogPath,
        String startDate,String endDate) throws Exception {
        if (isConnected) {
            // 存放日志文件的目录 为避免出现目录下有丙个或多个日志文件的情况 创建以设备IP命名的目录
            String path = tempLogPath + File.separator + dolbyLogPath + File.separator + ip;
            File logPath = new File(path);
            // 如果目录已存在就清理目录
            if (logPath.exists() && logPath.isDirectory()) {
                FileUtil.deleteFolder(path);
            }
            // 如果目录不存在 或刚执行过上面的清理目录 就创建目录
            // 如果目录创建失败 输出error日志
            if (!logPath.mkdirs()) {
                logger.error("DolbyServer getAuditLog occur error, cannot create log path : {}",
                    path);
                return null;
            }
            // 目录要有写权限
            logPath.setWritable(true, false);
            String result = systemManagement
                .getauditlogs(ip, ftpIp, user, pass, dolbyLogPath + File.separator + ip,
                    startDate, Constants.startTime, endDate, Constants.endTime, "", "");
            if ("ok".equals(result))
                return processFtpFile(path);
        }
        return null;
    }

    @Override
    public PlaybackStatus getPlaybackStatus() throws Exception {
        if (isConnected) {
            String xml = playbackControl.getplaybackstatus(ip);
//            logger.info("DolbyServer getPlaybackStatus serverIP : {}, result : {}", ip, xml);
            xml = xml.replace("&", "&amp;");
            PlaybackStatusResponse response = new PlaybackStatusResponse(xml);
            if (response.status() == Response.OK)
                return response.getPlaybackStatus();
        }
        return null;
    }

    /**
     * 从ftp特定目录获取日志文件
     *
     * @return
     */
    private String processFtpFile(String path) throws DocumentException {
        File logPath = new File(path);
        if (logPath.exists()) {
            File[] files = logPath.listFiles();
            if (null == files) {
                logger.error("processFtpFile error, files is null");
                return null;
            }
            // 存放日志的目录下没有日志文件
            if (0 == files.length) {
                logger.error("processFtpFile error, files is empty");
                return null;
            }
            // 存放日志的目录下有且只有一个日志文件时 对日志文件解析
            if (1 == files.length) {
                File xmlFile = files[0];
                SAXReader reader = new SAXReader();
                Document document = reader.read(xmlFile);
                return document.asXML();
            }
            // 存放日志的目录下有两个或多个日志文件
            logger.error("processFtpFile error, files length : {} > 2", files.length);
            throw new RuntimeException("processFtpFile error : " + Arrays.toString(files));
        }
        logger.error("processFtpFile error, logPath not exists : {}", path);
        return null;
    }
}
