package com.miaozhen.device.support.christienew.data;

/*!*******************************************************************************************************************
 * FILENAME: KlvMessage.java
 * CREATED:  2011/02/25
 *
 * Copyright (c) 2011 Christie Digital Canada Inc. All rights reserved.
 *
 * This is a class for dealing with CEP Messages
 *
 *********************************************************************************************************************/
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.util.*;


/**
 * Class for wrapping our KLVMessage 
 */
public class KlvMessage
{

   //Command set definitions
   public static final byte CMD_UNKNOWN =  ((byte)0x00);
   /*public static final byte CMD_Set = ((byte)0x01);
   public static final byte CMD_Get =((byte)0x02);
   public static final byte CMD_List=((byte)0x03);
   public static final byte CMD_Add=((byte)0x04);
   public static final byte CMD_Delete=((byte)0x05);
   public static final byte CMD_Update=((byte)0x06);
   public static final byte CMD_NotifyList=((byte)0x07);
   public static final byte CMD_NotifyUpdate=((byte)0x08);
   public static final byte CMD_NotifyGet=((byte)0x09);*/
        
    public static final int HeaderSize = 18;
    /// <summary> The size of the UL HEADER </summary>
    public static final int ULKeySize = 18;
    /// <summary> The size of a payloads Property name or key.</summary>
    public static final  int NameSize = 4;
    /// <summary> The size of the size field </summary>
    public static final int BerSize = 4;

    /// <summary> The Position of the command byte </summary>
    public static final int CmdPos = 6;
    /// <summary> The Position of the ObjectKey </summary>
    public static final int ObjKeyPos = 7;
    /// <summary>The Position of the Error bit </summary>
    public static final int ErrorPos = 9;
    
    public static final int Destination = 11;
    
    /// <summary> The Position of the SourceID byte </summary>
    public static final int SourceIDPos = 12;
    /// <summary> The Position of the Security Level byte </summary>
    public static final int SecurityLevelPos = 13;
    /// <summary> The Position of the CRC16 bytes </summary>
    public static final int CRC16Pos = 14;
    /// <summary> The Position of the Sequence number byte </summary>
    public static final int SequenceNumberPos = 16;
    /// <summary> The Position of the Reserved bytes </summary>
    public static final int ReservedPos = 17;

    //used for calculating CRC
    static int[] table = {
            0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
            0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
            0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
            0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
            0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
            0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
            0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
            0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
            0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
            0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
            0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
            0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
            0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
            0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
            0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
            0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
            0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
            0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
            0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
            0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
            0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
            0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
            0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
            0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
            0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
            0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
            0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
            0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
            0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
            0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
            0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
            0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040,
        };

    /// The error state of the message
    public short ErrorState;
    /// The number of bytes currently stored in the payload of this message
    private int m_byteCount = 0;
    /// Message Sequence Number
    private byte m_sequenceNumber = 0;
    ///  The Source ID of this message 0 is no particular source 
    private byte m_sourceID = 0;
    
    private byte destination = 0;
    ///  The CRC16 of this message
    private short m_CRC16 = 0;
    /// The Security level of this message
    private byte m_securityLevel = 0;
    /// Object Key Array
    private short m_objectKey = 0;
    /// A <i>DateTime</i> object representing the start of Posix time (Jan. 1, 1970).
    public static Date posixTimeOrigin = new Date(1970, 1, 1, 0, 0, 0);
    /// The series of bytes that denote the start of a KLV message. Includes the version number fields.
    private final static byte[] klvMessageUniversalLabel = { 0x7b, 0x02, 0x34, 0x2b, 0x02, 0x00 };
    /// Command ID
    public byte Command = CMD_UNKNOWN;
    /// Storage of message data in bytes.
    private Map<Integer,byte[]> data = new HashMap<Integer,byte[]>();

    /**
     * Returns the objectkey of this message
     * @return
     */
    public short GetObjKey(){return m_objectKey;}
    /**
     * Sets the object key of this message
     * @param key
     */
    public void SetObjKey(short key){m_objectKey=key;}

	/**
	 * Checks to see if a key is present in the message
	 * @param name
	 * @return
	 */
    public boolean KeyExists(int name)
    {
       return data.containsKey(name);

    }

    /**
     * 	return a list of keys for this message
     * @return
     */
	public int[] GetMessageKeys(){
    	int i=0;
        int[] retVal = new int[data.keySet().size()];
        Iterator<Integer> iter = data.keySet().iterator();
        while(iter.hasNext()){
        	retVal[i++]= iter.next().intValue();
        }
        return retVal;
    }

    /**
     * Default constructor
     */
    public KlvMessage()
    {

    }
    /**
     * Returns the source id
     * @return
     */
    public byte GetSourceId()
    {
        return m_sourceID;
    }

    /**
     * Sets the source ID for this message
     * @param id
     */
    public void SetSourceId(byte id)
    {
        m_sourceID = id;
    }

    /**
     * Special value for source ID used internally
     */
    public final byte DontTransmitSourceID = Byte.MAX_VALUE;


    /**
     * Sets the CRC16 Value of the message
     * @param crc16
     */
    public void SetCRC16(short crc16)
    {
        m_CRC16 = crc16;
    }

	/**
	 *
	 * @return the byte permission level
	 */
    public byte GetPermission()
    {
        return m_securityLevel;
    }

    /**
     * Returns the security level of the message, this is set by the system when message are received. Any value here
     * will be changed to what the current user level is for the thread the message is recieved on.
     * @param
     */

    public void SetSecurityLevel(byte level)
    {
        m_securityLevel = level;
    }

    /**
     * Copies this message and returns a new instance
     * @return
     */
    public KlvMessage Copy()
    {
        KlvMessage cmsg = new KlvMessage();
        cmsg.Command = Command;
        cmsg.m_sourceID = m_sourceID;
        cmsg.m_securityLevel = m_securityLevel;
        cmsg.m_objectKey = m_objectKey;
        cmsg.ErrorState = ErrorState;
        cmsg.m_sequenceNumber = m_sequenceNumber;
        int[] keys = GetMessageKeys();
        for (int k = 0; k < keys.length; k++)
        {
            cmsg.SetProperty(keys[k], GetProperty(keys[k]));
        }
        return cmsg;
    }

    /**
     * Compares the set of keys between this message and the supplied message for equivalence.
     * @param right
     * @return
     */
    public boolean KeysCompareEqual(KlvMessage right)
    {
        boolean retVal = false;

        if (right != null)
        {
            int[] leftKeys = GetMessageKeys();
            int[] rightKeys = right.GetMessageKeys();
            if (leftKeys.length == rightKeys.length)
            {
                retVal = true;
                for (int i = 0; i < leftKeys.length; ++i)
                {
                    if (leftKeys[i] != rightKeys[i])
                    {
                        retVal = false;
                        break;
                    }
                }
            }
        }
        return retVal;
    }

    /**
     * Adds a raw property to the message
     * @param name The name of the property that is being added
     * @param bytes The byte[] of data to be added
     * @return True if successful, false if the <i>bytes</i> parameter is null.
     */
    public boolean SetProperty(int name, byte[] bytes)
    {
        if (bytes != null)
        {
            if (data.containsKey(name))
            {
                // The name exists. Replace the value associated with the name.
                // The replace is implemented as a remove followed by add.
                // The Remove method adjusts the byte count and removes the old value.
                //
                Remove(name);
            }
            m_byteCount += bytes.length + NameSize + Berlength(bytes.length)+1; // Each Property has key,length,value
            data.put(name, bytes.clone());
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * Returns the Raw bytes version of the property by name
     * @param name The name of the property to return
     * @return The byte[] of the data for the specified property
     */
    public byte[] GetProperty(int name)
    {
        return (byte[])data.get(name);
    }

    /**
     * Sets bool data on the message
     * @param name The name of the property to be added
     * @param value The value to be set
     * @return True if successful, False if not
     */
    public boolean SetBool(int name, boolean value)
    {
        byte[] bytes = new byte[1];
        if (value)
        {
            bytes[0] = 1;
        }
        else
        {
            bytes[0] = 0;
        }
        SetProperty(name, bytes);
        return true;
    }

    /**
     * Gets bool data from the message
     * @param name The name of the property to be returned
     * @return The boolean value of the property with specifed name.
     * @throws MissingFieldException
     */
    public boolean GetBool(int name)throws MissingFieldException
    {
        try
        {
            // Default return is true
            boolean bReturn = true;

            // Get the data
            byte[] bytes = (byte[])data.get(name);

            // If the value of the byte is zero, then change return to false
            if (bytes.length==0 ||bytes[0] == 0)
            {
                bReturn = false;
            }
            return bReturn;
        }
        catch (Exception e)
        {
           throw new MissingFieldException(e.toString() + " key name: " + name);

        }
    }

    /**
     * Sets byte data on the message
     * @param name The name of the property to be added
     * @param value The value to be set
     * @return True if successful, False if not
     */
    public boolean SetByte(int name, Byte value)
    {
        byte[] bytes = new byte[1];
        bytes[0] = value;

        this.SetProperty(name, bytes);
        return true;
    }
    /**
     * Gets byte data from the message
     * @param name The name of the property to be returned
     * @return The byte value of the specified property name
     * @throws MissingFieldException
     */
    public Byte GetByte(int name)throws MissingFieldException
    {
        try
        {
            byte[] bytes = (byte[])data.get(name);
            return bytes[0];
        }
        catch (Exception e)
        {
            throw new MissingFieldException(e.toString() + " key name: " + name);

        }
    }

    /**
     * Sets an Int32 value on the message
     * @param name The name of the property to be added
     * @param value The value to be set
     * @return True if successful, False if not
     */
    public boolean SetInt32(int name, int value)
    {
        byte[] b = KlvMessage.intToByteArray(value);
        this.SetProperty(name, b);
        return true;
    }

    /**
     * Gets an Int32 value from the message
     * @param name The name of the property to be returned
     * @return The Int32 value to be returned.
     * @throws MissingFieldException
     */
    public int GetInt32(int name) throws MissingFieldException
    {
        try
        {
            byte[] obj = (byte[])data.get(name);

            int ret = KlvMessage.ByteArrayToInt(obj);

            return ret;
        }
        catch (IllegalArgumentException ae)
        {

            throw new MissingFieldException("GetInt32 asked for missing element " + name, ae);
        }
        catch (Exception ae){
        	throw new MissingFieldException("GetInt32 asked for missing element " + name, ae);
        }
    }

    /**
     * Sets an Int64 value on the message
     * @param name The name of the property to be added
     * @param value The value to be set
     * @return True if successful, False if not
     */
    public boolean SetInt64(int name, long value)
    {
        byte[] b = KlvMessage.longToByteArray(value);
        this.SetProperty(name, b);
        return true;
    }


    /**
     * Gets an Int64 value from the message.
     * @param name The name of the property to be returned
     * @return The Int64 value to be returned.
     * @throws MissingFieldException
     */
    public long GetInt64(int name)throws MissingFieldException
    {
        try
        {
            byte[] obj = data.get(name);
            return KlvMessage.ByteArrayTolong(obj);

        }
        catch (IllegalArgumentException e)
        {
            throw new MissingFieldException(e.toString());
        }
        catch (Exception e)
        {
            throw new MissingFieldException(e.toString());
        }

    }


    /**
     * Sets an float value on the message
     * @param name The name of the property to be added
     * @param value The value to be set
     * @return True if successful, False if not
     */
    public boolean SetFloat(int name, float value)
    {
    	  	
    	
        byte[] b = KlvMessage.FloatToBytes(value); 
        this.SetProperty(name, b);
        return true;

    }
    
    /**
     * Gets an float value from the message
     * @param name The name of the property to be returned
     * @return The float value to be returned
     * @throws MissingFieldException
     */
    public float GetFloat(int name) throws MissingFieldException
    {
        try
        {
            byte[] obj = data.get(name);
            return KlvMessage.ByteArrayToFloat(obj);

        }
        catch (Exception e)
        {
            throw new MissingFieldException(e.toString());
        }
    }
 
    /**
     * Sets a unicode string value on the message in format UTF-16LE
     * @param name The name of the property to be added
     * @param value The value to be set
     * @return True if successful, otherwise false if value is null or cannot encode the characters.
     */
    public boolean SetString(int name, String value)
    {
        if (value != null)
        {
        	
                try {
					return SetProperty(name, value.getBytes("UTF-16LE"));
				} catch (UnsupportedEncodingException e) {
					
				}
          
        }
        return false;
    }
  
    /**
     * Get a unicode string value from the message. Format UTF-16LE
     * @param name The name of the property to be returned.
     * @return Returns the string at the specified property. Returns an empty string
     * if the property does not exist or if it cannot convert the data to a string.
     */
    public String GetString(int name)
    {
        byte[] bytes = (byte[])data.get(name);
        if (bytes != null)
        {
                try {
					return new String(bytes,"UTF-16LE");
				} catch (Exception e) {
				}
           
        }
        return "";
    }
    
    /**
     * Gets a specified UTF8 String     
     * @param name The key of the property to be returned
     * @return
     */
    public String GetStringUTF8(int name)
    {
        byte[] bytes = (byte[])data.get(name);
        if (bytes != null)
        {
                try {
					return new String(bytes,"UTF-8");
				} catch (Exception e) {
				}
           
        }
        return "";
    }
    

    /**
     * Sets an ASCII string value on the message.
     * @param name The name of the property to be added
     * @param value The value to be set.
     * @return True if successful, otherwise false if value is null or cannot encode characters to ASCII.
     */
    public boolean SetAsciiString(int name, String value)
    {
        if (value != null)
        {
                try {
					return SetProperty(name, value.getBytes("US-ASCII"));
				} catch (UnsupportedEncodingException e) {
				
				}
        }
        return false;
    }
   
    /**
     * Gets an ASCII string value from the message.
     * @param name The name of the property to be returned.
     * @return Returns the string at the specified property. Returns an empty string
     * if the property does not exist or if it cannot convert the data to a string.
     */
    public String GetAsciiString(int name)
    {
        byte[] bytes = (byte[])data.get(name);
        if (bytes != null)
        {
                try {
					return new String(bytes,"US-ASCII");
				} catch (UnsupportedEncodingException e) {
				
				}
        }
        return "";
    }
    
    /**
     * Sets a Message property value
     * 
     * @param name The name of the property to be added
     * @param msg The value to be set
     * @return True if successful, False if not
     */
    public boolean SetMessage(int name, KlvMessage msg)
    {
        byte[] bytes = new byte[msg.m_byteCount];

        msg.SerializeProperties(bytes, 0);
        return SetProperty(name, bytes);
    }
    

    /**
     * @return the length of the message 
     */
    public int length()
    {
    	return m_byteCount + HeaderSize + Berlength(m_byteCount);
    }


    /**
     * 
     * @return bytes of the message with crc16 attached
     */
    public byte[] ToBytesWithCRC16()
    {
        byte[] dataBytes = ToBytes();
        short crc = CalculateCRC(dataBytes);
      
        byte[] crcBytes = shortToByteArray(crc);

        dataBytes[CRC16Pos] = (byte)(crcBytes[0]& 0xFF);
        dataBytes[CRC16Pos + 1] = (byte)(crcBytes[1]& 0xFF);

        return dataBytes;
    }

    /**
     * Calculate the CRC value of the array of bytes
     * @param bytes
     * @return
     */
    public static short CalculateCRC(byte[] bytes){
        int crc = 0x0000;
        for (byte b : bytes) {
            crc = (crc >>> 8) ^ table[(crc ^ b) & 0xff];
        }
    	return (short)crc;    		      
    }
    
   /**
    *  Encode the KlvMessage to a byte array
    * @return byte array if Successful, null Otherwise
    */
    public byte[] ToBytes()
    {
        int len = length();
        byte[] buffer = new byte[len];

        int buffer_index = 0;
        //----------------------------------------
        // Set the UL label
       System.arraycopy(klvMessageUniversalLabel,0,buffer, buffer_index,klvMessageUniversalLabel.length);

        //----------------------------------------
        // Set the command
        buffer[CmdPos] = (byte)Command;

        //----------------------------------------
        // Set the object key
        buffer_index = ObjKeyPos;
       // buffer[buffer_index] = m_objectKey;
        //----------------------------------------
        // Set the special bits
        buffer[SecurityLevelPos] = m_securityLevel;

        byte[] errorBytes = shortToByteArray(ErrorState);
        buffer[ErrorPos]=errorBytes[0];
        buffer[ErrorPos+1]=errorBytes[1];
        buffer[SourceIDPos] = m_sourceID;
        buffer[SequenceNumberPos] = m_sequenceNumber;

        // Set the Payload Size
        byte[] psize = EncodelengthToBer(m_byteCount);
        
        
        //System.out.println("BER Length "+m_byteCount + " : " + Arrays.toString(psize));
        System.arraycopy(psize,0,buffer,HeaderSize,psize.length);
        SerializeProperties(buffer, HeaderSize + psize.length);
        return buffer;
    }
 
    /**
     * Serializes a sub-message into an array of bytes.
     * @return byte array representation of the KLV message payload.
     */
    public byte[] PayloadToBytes()
    {
        byte[] buffer = new byte[m_byteCount];
        SerializeProperties(buffer, 0);
        return buffer;
    }
    
    /**
     * Serializes a sub-message into an array of bytes.
     * @param buffer The buffer that serves as the destination of the serialized data.
     * @param index The index into the buffer at which at start serialization.
     */
    private void SerializeProperties(byte[] buffer, int index)
    {
        int[] keys = GetMessageKeys();
        for (int k=0;k<keys.length; k++)
        {
            // Get the value by name
            //
            byte[] value = (byte[])data.get(keys[k]);

            // Add the name
            //
            byte[] bname = KlvMessage.intToByteArray(keys[k]);
            System.arraycopy(bname, 0, buffer, index, bname.length);
            index += bname.length;

            // Add the length
            byte[] blen = EncodelengthToBer(value.length);
            //System.out.println("Serialize Properties " + Arrays.toString(blen));
            System.arraycopy(blen, 0, buffer, index, blen.length);
            index += blen.length;

            // Add the value
            System.arraycopy(value, 0, buffer, index,value.length);
            index += value.length;
        }
     }
    
    /**
     * Decodes the KlvMessage from a byte array.
     * @param buffer byte array to decode
     * @return KlvMessage if okay else null
     */
    public static KlvMessage FromBytes(byte[] buffer)
    {
        //Check to see if header is valid
        if (CheckKlvMessageHeaderUL(buffer) == false)
        {
            //Message is not valid;
            return null;
        }

        KlvMessage msg = new KlvMessage();

        ////////////////////////////////////
        // Decode Header
        ////////////////////////////////////
        // Get Command
        msg.Command = buffer[CmdPos];

        // Get OID
        msg.m_objectKey = buffer[ObjKeyPos];
        ////////////////////////////
        // Get Special Bits (3 Bytes left)
        ////////////////////////////

        msg.SetSourceId(buffer[SourceIDPos]);
        
        byte[] errorBytes = new byte[2];
        System.arraycopy(buffer, ErrorPos, errorBytes, 0, 2);
        msg.ErrorState = KlvMessage.ByteArrayToShort(errorBytes);
        msg.SetSecurityLevel(buffer[SecurityLevelPos]);

        // read the Sequence number
        msg.m_sequenceNumber = buffer[SequenceNumberPos];

        // Get the total length of all properties and deserialize just that many bytes,
        // ignoring any additional bytes if the buffer is bigger.
        
        BERLen ber = KlvMessage.DecodelengthFromBer(buffer, ULKeySize);
        msg.DeserializePayload(buffer, HeaderSize + ber.BERLength, ber.PropertyLength);

        return msg;
    }

    /**
     *  Decodes a sub-message from a byte array.
     * @param buffer The array of bytes to decode
     * @param index The start index within the buffer
     * @param length The number of bytes to decode.
     */
    public void DeserializePayload(byte[] buffer, int index, int length)
    {
        int end = index + length;

        while (index < end && index+4<buffer.length)
        {
            // Get the property name
            try
            {
            	
	            	byte[] b = new byte[4];
	            	System.arraycopy(buffer,index,b,0,4);
	                int name = KlvMessage.ByteArrayToInt(b);
	                index += NameSize;
	
	                // Get the property length
	                int propsize;
	                BERLen ber;
	                 
	                ber= KlvMessage.DecodelengthFromBer(buffer, index);
	                int plengthlength = ber.BERLength;
	                propsize=ber.PropertyLength;
	                index += plengthlength;
	
	                // Get the value 
	                byte[] propbytes = new byte[propsize];
	                System.arraycopy(buffer, index, propbytes, 0, propsize);
	                index += propsize;
	
	                // Add the property to the message
	                SetProperty(name, propbytes);
            	
            }
            catch (IllegalArgumentException ae)
            {
                index = length;
            }
        }
    }

    /**
     * This method searches for the ULKey in the given array of bytes.
     * If the key or part of the key is found in the incoming bytes return 
     * the position the key was found. 
     * @param bytes The bytes to be searched
     * @return The start position of the ULKey, or -1 if it was not found
     */
    public static int FindUL(byte[] bytes)
    {
        return FindUL(bytes, 0, bytes.length);
    }

    /**
     * This method searches for the ULKey in the given array of bytes.
     * If the key or part of the key is found in the incoming bytes return 
     * the position the key was found. 
     * @param bytes The bytes to be searched
     * @param offset The number of bytes to search 
     * @param length length">The number of bytes to search
     * @return The start position of the ULKey, or less then 0 (zero) if it was not found
     */
    public static int FindUL(byte[] bytes, int offset, int length)
    {
        //  int searchResult = Array.BinarySearch(bytes, offset, length, KlvMessage.klvMessageUniversalLabel, null);
        boolean bFound = false;
        int firstKey = KlvMessage.klvMessageUniversalLabel[0];
        int pos;

        // Compute the index of the last character to check.
        // There is no need to scan the last few bytes for the start of a header so we subtract off the UL length.
        //
        int lastIndex = offset + length - KlvMessage.klvMessageUniversalLabel.length;

        // Loop from the start index to the end of the array
        for (pos = offset; pos <= lastIndex; pos++)
        {
            // If the first key has been found then process
            if (firstKey == bytes[pos])
            {
                bFound = true;

                // Loop through the buffer
                for (int i = 1; i < KlvMessage.klvMessageUniversalLabel.length; i++)
                {
                    // If the byte and the key do not match
                    if (bytes[pos + i] != KlvMessage.klvMessageUniversalLabel[i])
                    {
                        bFound = false;
                        break;
                    }
                }
                // If the key found then break
                if (bFound)
                {
                    break;
                }
            }
        }
        if (bFound)
        {
            return pos;
        }
        else
        {
            return -1;
        }

    }

    /**
     * Checks for the klvMessage header at the start of a buffer.
     * @param buffer The byte array to check.
     * @return Returns true if the start of the buffer contains the universal label.
     */
    private static boolean CheckKlvMessageHeaderUL(byte[] buffer)
    {
        boolean areEqual = false;
        if (buffer.length >= klvMessageUniversalLabel.length)
        {
            areEqual = true;
            for (int index = 0; index < klvMessageUniversalLabel.length; index++)
            {
                if (buffer[index] != klvMessageUniversalLabel[index])
                {
                    areEqual = false;
                    break;
                }
            }
        }
        return areEqual;
    }


    /**
     * Decode length from BER Bytes
     * @param buffer Bytes to decode.
     * @param index Index into byte array.
     * @return true if successful, false otherwise.
     */
    public static BERLen DecodelengthFromBer(byte[] buffer, int index)
    {
        // Default 
    	BERLen result = new BERLen();
    	
        int temp_len = 0;
        int berlength = 0;
        int bufferSize = buffer.length;
        // If the index is not in the buffer the return zero
        if (bufferSize > index)
        {
        	
            if ((buffer[index]&0xff) >= 0x80)
            {
                berlength =(buffer[index]&0xff) & 0x7f;
                //sanity check
                if (berlength > 3 || berlength == 0)
                {
                    berlength = 0;  // our error flag
                }
                else
                {
                    index++;
                    for (int i = 0; i < berlength; i++)
                    {
                    	int value = buffer[index++]&0xff;
                    	int value2=(temp_len << 8);
                        temp_len = value2 + value;
                    }
                    ++berlength;   // Increment to include the initial byte in the count.
                }
            }
            else
            {
            	 
                temp_len = buffer[index]&0xff;
                berlength = 1;  //the base case, one byte, 127 or under.
            }
        }
        result.BERLength=berlength;
        result.PropertyLength=temp_len;
        return result;
    }


    /**
     * Encodes a length value to BER.
     * @param valueToEncode
     * @return Returns a byte array with the encoded length. On failure, the byte array is empty.
     */
    public static byte[] EncodelengthToBer(int valueToEncode)
    {
        byte[] buffer;
        if (valueToEncode < 128)
        {
            buffer = new byte[1];
            buffer[0] = (byte)valueToEncode;
        }
        else if (valueToEncode < 0x100)
        {
            buffer = new byte[2];
            buffer[0] = (byte)0x80 | (byte)0x01;   // 0x01 means next one byte contains length
            buffer[1] = (byte)valueToEncode;
        }
        else if (valueToEncode < 0x10000)
        {
            buffer = new byte[3];
            buffer[0] = (byte)0x80 | (byte)0x02;   // 0x02 means next two bytes contain length
            buffer[1] = (byte)(0xFF & (valueToEncode >> 8));
            buffer[2] = (byte)(0xFF & valueToEncode);
        }
        else if (valueToEncode < 0x1000000)
        {
            buffer = new byte[4];
            buffer[0] = (byte)0x80 | (byte)0x03;   // 0x03 means next three bytes contain length
            buffer[1] = (byte)(0xFF & (valueToEncode >> 16));
            buffer[2] = (byte)(0xFF & (valueToEncode >> 8));
            buffer[3] = (byte)(0xFF & valueToEncode);
        }
        else
        {
            buffer = new byte[0];   // empty buffer
            //error, over 1MB not supported
        }

        return buffer;
    }
    /**
     * 
     * did not definitely say which message is less or more, by the old compare algorithms
     * @return 0 if the same, not 0 if different
     */
    public int CompareTo( KlvMessage other )
    {
        int result = 0;

        if (other == null)
        {
            result = 1;
        }
        else
        {
            if ( this.ErrorState != other.ErrorState ||
                this.GetObjKey() != other.GetObjKey() ||
                this.Command != other.Command ||
                this.length() != other.length() )
            {
                result = 1;
            }
            else
            {
                int[] keys = this.GetMessageKeys();
                int[] otherKeys = other.GetMessageKeys();
                result = keys.length - otherKeys.length;
                if (result == 0)
                {
                    for (int myKey : keys)
                    {
                        if (other.KeyExists( myKey ))
                        {
                            byte[] myValue = this.GetProperty( myKey );
                            byte[] otherValue = other.GetProperty( myKey );
                            if (!Arrays.equals( myValue, otherValue ))
                            {
                                result = 1;
                            }
                        }
                        else
                        {
                            result = 1;
                            break;
                        }
                    }
                }
            }
        }
        return result;
    }
    
    /**
     * Clears all properties from the message
     */
    public void Clear()
    {
        m_byteCount = 0;
        data.clear();
    }

    /**
     * Removes a property from a message
     * @param name
     */
    public void Remove(int name)
    {
        byte[] oldbytes = (byte[])data.get(name);
        m_byteCount -= oldbytes.length + NameSize + Berlength(oldbytes.length)+1;
        data.remove(name);
    }
    
    /**
     *  Computes the number of bytes required to store the BER length.
     * @param datalength
     * @return Returns the length of the BER in bytes
     */
    private static int Berlength(int datalength)
    {
        int length = 4;       // Maximum allowed length.

        if (datalength < 128)
        {
            length = 1;
        }
        else if (datalength < 0x100)
        {
            length = 2;
        }
        else if (datalength < 0x10000)
        {
            length = 3;
        }
        return length;
    }
    
    /**
     * Convert short to bytearray
     * @param value
     * @return
     */
    public static final byte[] shortToByteArray(short value) {
    	
    	
    	return new byte[]{(byte)(value ),(byte)(0xFF&value >> 8)};
    }

    /**
     * Convert Int to byte array
     * @param value
     * @return
     */
    public static final byte[] intToByteArray(int value) {
    	
    	
    	return new byte[]{(byte)(value ),(byte)(0xFF&value >> 8),(byte)(0xFF&value >> 16),(byte)(0xFF&value>> 24)};
    }

    /**
     * Convert ByteArray to show
     * @param data
     * @return
     */
    public static final short ByteArrayToShort(byte[] data) {
    	
    	return (short)( (data[0]&0xFF) |((data[1]& 0xFF)<<8));
    	
    }
    
    /**
     * Convert ByteArray to Int
     * @param data
     * @return
     */
    public static final int ByteArrayToInt(byte[] data) {
    	
    	return (int)( (data[0]&0xFF) |((data[1]& 0xFF)<<8)|((data[2]&0xFF)<<16)| ((data[3]&0xFF)<<24));
    	
    }
    
    /**
     * Converts bytes to bytes
     * @param value
     * @return
     */
    public static final byte[] longToByteArray(long value) {
    	return new byte[]{(byte)(value ),(byte)(0xFF&value >> 8),(byte)(0xFF&value >> 16),(byte)(0xFF&value>> 24),(byte)(0xFF&value >> 32), (byte)(0xFF&value >> 40 ),(byte)(0xFF&(value >> 48) ),(byte)(0xFF&(value >> 56)) };
    }
    
    /**
     * Converts bytes to long
     * @param data
     * @return
     */
    public static final long ByteArrayTolong(byte[] data) {
    	long value = (data[0]&0xFF);
    			value |= (long)(data[1]&0xFF)<<8; 
    			value |=  (long)(data[2]&0xFF)<<16;
    			value |= (long)(data[3]&0xFF)<<24;
    			
    			value |= (long)(data[4]&0xFF)<<32;
    			value|= (long)(data[5]&0xFF)<<40;
    			value|= (long)(data[6]&0xFF)<<48;
    			value|= (long)(data[7]&0xFF)<<56 ;
    	return value;
    	
    }
    
    /**
     * Converts bytes to float 
     * @param data
     * @return
     */
    public static final float ByteArrayToFloat(byte[] data){
    	 ByteBuffer buf = ByteBuffer.wrap(data);
    	 return buf.getFloat();
    }

    /**
     * Converts float to bytes
     * @param value
     * @return
     */
    public static final byte[] FloatToBytes(float value){
    	
    	 ByteBuffer buf = ByteBuffer.allocate(4);
    	 buf.putFloat(value);
    	 return buf.array();
    	
    }    
}
