package com.miaozhen.task.manager;

import static org.quartz.JobKey.jobKey;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;

import com.miaozhen.task.JobInterface;
import com.miaozhen.task.QuartzJobFactory;
import com.miaozhen.task.domain.ScheduleJob;
import com.miaozhen.task.job.ClearLogJob;
import com.miaozhen.task.job.GetCinemaInfoJob;
import com.miaozhen.task.job.ServerLogJob;
import org.quartz.*;
import org.quartz.DateBuilder.IntervalUnit;
import org.quartz.impl.StdSchedulerFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * 定时任务管理类
 *
 * @author fengtao
 */
@Component
public class QuartzManager {

    private final Logger log = LoggerFactory.getLogger(getClass());

    private Scheduler scheduler;

    private Map<String, ScheduleJob> jobMap = new HashMap<String, ScheduleJob>();

    private final SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");

    private SchedulerFactory schedulerFactory = new StdSchedulerFactory();

    public Scheduler getScheduler() {
        return scheduler;
    }

    @Autowired
    ServerLogJob serverLogJob;
    @Autowired
    GetCinemaInfoJob getCinemaInfoJob;
    @Autowired
    ClearLogJob clearLogJob;

    /**
     * 初始化jobMap
     *
     * @param
     */
    public void initJobMap() {
        createGetCinemaInfoJob();
        createServerLogJob();
        createClearLogJob();
    }

    /**
     * 初始化任务并启动任务
     */
    public void initTask() {
        log.info("QuartzManager: init Task begin");
        try {
            scheduler = schedulerFactory.getScheduler();
            Collection<ScheduleJob> jobList = jobMap.values();
            for (ScheduleJob job : jobList) {
                TriggerKey triggerKey = TriggerKey.triggerKey(job.getJobName(), job.getJobGroup());
                Trigger trigger = scheduler.getTrigger(triggerKey);
                if (null == trigger) {
                    addJob(QuartzJobFactory.class, job);
                } else {
                    modifyJob(job);
                }
            }

        } catch (SchedulerException e1) {
            log.error("init Task SchedulerException ", e1);
        }
        try {
            log.info("***************************QuartzManager is start**************************");
            // 启动定时任务
            scheduler.start();
        } catch (SchedulerException e) {
            log.error("initTask  scheduler.start() SchedulerException ", e);
        }
    }

    /**
     * 获取影院放映服务器播放日志
     *
     * @return
     */
    public ScheduleJob createServerLogJob() {
        ScheduleJob groupJob = new ScheduleJob();
        groupJob.setBeginTime(ScheduleJob.SERVERLOG_JOB_DEFAULT_BEGIN);
        groupJob.setJobId(ScheduleJob.SERVERLOG_JOB_ID);
        groupJob.setJobName(ScheduleJob.SERVERLOG_JOB_NAME);
        groupJob.setJobGroup(ScheduleJob.SERVERLOG_JOB_GROUP);
        groupJob.setJobStatus("1");
        groupJob.setNumber(0);
        groupJob.setJob(serverLogJob);
        groupJob.setIntervalUnit(IntervalUnit.MINUTE);
        groupJob.setIntervalTime(ScheduleJob.SERVERLOG_JOB_DEFAULT_INTERVAL);
        groupJob.setDesc("获取影院放映服务器播放日志");
        addJob(groupJob);
        return groupJob;
    }

    /**
     * 获取影院放映服务器配置信息
     *
     * @return
     */
    public ScheduleJob createGetCinemaInfoJob() {
        ScheduleJob groupJob = new ScheduleJob();
        groupJob.setBeginTime(ScheduleJob.GETCINEMAINFO_JOB_DEFAULT_BEGIN);
        groupJob.setJobId(ScheduleJob.GETCINEMAINFO_JOB_ID);
        groupJob.setJobName(ScheduleJob.GETCINEMAINFO_JOB_NAME);
        groupJob.setJobGroup(ScheduleJob.GETCINEMAINFO_JOB_GROUP);
        groupJob.setJobStatus("1");
        groupJob.setNumber(0);
        groupJob.setJob(getCinemaInfoJob);
        groupJob.setIntervalUnit(IntervalUnit.MINUTE);
        groupJob.setIntervalTime(ScheduleJob.GETCINEMAINFO_JOB_DEFAULT_INTERVAL);
        groupJob.setDesc("获取影院配置信息");
        addJob(groupJob);
        return groupJob;
    }

    /**
     * 日志定时清理
     *
     * @return
     */
    public ScheduleJob createClearLogJob() {
        ScheduleJob groupJob = new ScheduleJob();
        groupJob.setBeginTime(ScheduleJob.CLEARLOG_JOB_DEFAULT_BEGIN);
        groupJob.setJobId(ScheduleJob.CLEARLOG_JOB_ID);
        groupJob.setJobName(ScheduleJob.CLEARLOG_JOB_NAME);
        groupJob.setJobGroup(ScheduleJob.CLEARLOG_JOB_GROUP);
        groupJob.setJobStatus("1");
        groupJob.setNumber(0);
        groupJob.setJob(clearLogJob);
        groupJob.setIntervalUnit(IntervalUnit.MINUTE);
        groupJob.setIntervalTime(ScheduleJob.CLEARLOG_JOB_DEFAULT_INTERVAL);
        groupJob.setDesc("日志定时清理");
        addJob(groupJob);
        return groupJob;
    }

    public void addJob(ScheduleJob scheduleJob) {
        jobMap.put(scheduleJob.getJobGroup() + "_" + scheduleJob.getJobName(), scheduleJob);
    }

    /**
     * @param clsImplJob  传入实现JOB的类 Class<? extends Job> cls = ExampleJob.class;
     * @param scheduleJob 工作任务
     * @throws SchedulerException Scheduler操作的异常
     * @throws ParseException     setCronExpression设置定时执行时间的异常
     * @Title: addJob
     * @Description: 添加任务
     */
    public void addJob(Class<? extends Job> clsImplJob, ScheduleJob scheduleJob) throws SchedulerException {
        log.info("QuartzManager:addJob {}", scheduleJob.getJobName());
        JobInterface job = scheduleJob.getJob();
        JobDetail myJob = JobBuilder.newJob(clsImplJob)
                .withIdentity(scheduleJob.getJobName(), scheduleJob.getJobGroup()).build();
        myJob.getJobDataMap().put(QuartzJobFactory.SCHEDULEJOB_KEY, scheduleJob);
        Date startTime = DateBuilder.futureDate(scheduleJob.getBeginTime(), scheduleJob.getIntervalUnit());
        // 通过工厂创建一个调度者
        SimpleTrigger trigger = TriggerBuilder.newTrigger()
                .withIdentity(scheduleJob.getJobName(), scheduleJob.getJobGroup()).startAt(startTime)
                .withSchedule(SimpleScheduleBuilder.simpleSchedule()
                        .withIntervalInSeconds(scheduleJob.getIntervalTime()).repeatForever())
                .build();
        Date ft = scheduler.scheduleJob(myJob, trigger);
        log.info("Task name：{} In： {} will be executed. And repeat：{} times，Time interval {} second",
                myJob.getKey().getName(), dateFormat.format(ft), trigger.getRepeatCount(),
                trigger.getRepeatInterval() / 1000);

    }

    /**
     * 修改任务
     *
     * @param scheduleJob
     * @throws SchedulerException
     */
    public void modifyJob(ScheduleJob scheduleJob) throws SchedulerException {
        log.info("QuartzManager:modifyJob {}", scheduleJob.getJobName());
        Date ft = null;
        String jobName = scheduleJob.getJobName();
        String jobGroupName = scheduleJob.getJobGroup();
        TriggerKey triggerKey = TriggerKey.triggerKey(jobName, jobGroupName);
        JobKey jobKey = JobKey.jobKey(jobName, jobGroupName);
        Date startTime = DateBuilder.futureDate(scheduleJob.getBeginTime(), scheduleJob.getIntervalUnit());
        JobDetail jobDetail = null;
        SimpleTrigger trigger = (SimpleTrigger) scheduler.getTrigger(triggerKey);
        if (trigger == null) {
            jobDetail = JobBuilder.newJob(QuartzJobFactory.class)
                    .withIdentity(scheduleJob.getJobName(), scheduleJob.getJobGroup()).build();
            jobDetail.getJobDataMap().put("scheduleJob", scheduleJob);
            trigger = trigger.getTriggerBuilder().withIdentity(scheduleJob.getJobName(), scheduleJob.getJobGroup())
                    .startAt(startTime).withSchedule(SimpleScheduleBuilder.simpleSchedule()
                            .withIntervalInSeconds(scheduleJob.getIntervalTime()).repeatForever())
                    .build();
            ft = scheduler.scheduleJob(jobDetail, trigger);
        } else {
            jobDetail = scheduler.getJobDetail(jobKey);
            // 按新的cronExpression表达式重新构建trigger
            trigger = trigger.getTriggerBuilder().withIdentity(scheduleJob.getJobName(), scheduleJob.getJobGroup())
                    .startAt(startTime).withSchedule(SimpleScheduleBuilder.simpleSchedule()
                            .withIntervalInSeconds(scheduleJob.getIntervalTime()).repeatForever())
                    .build();
            // 按新的trigger重新设置job执行
            ft = scheduler.rescheduleJob(triggerKey, trigger);
            log.info("Task name：{} In： {} will be executed. And repeat：{} times，Time interval {} second",
                    jobDetail.getKey().getName(), dateFormat.format(ft), trigger.getRepeatCount(),
                    trigger.getRepeatInterval() / 1000);
        }

    }

    /**
     * 删除任务
     *
     * @param scheduleJob
     * @throws SchedulerException
     */
    public void removeJob(ScheduleJob scheduleJob) throws SchedulerException {
        TriggerKey triggerKey = TriggerKey.triggerKey(scheduleJob.getJobName(), scheduleJob.getJobGroup());
        JobKey jobKey = jobKey(scheduleJob.getJobName(), scheduleJob.getJobGroup());
        scheduler.pauseTrigger(triggerKey);
        scheduler.unscheduleJob(triggerKey);
        scheduler.deleteJob(jobKey);
    }

}
